/*
 *  	int Forthish: "Everything's an int!"
 * 
 *	All-in-one lexer, parser, code generator (kind of)
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <allegro.h>
#include "bytecode.h"


/*-------------------------------------------------------------------
 	A "works for most cases" lexer
 -------------------------------------------------------------------*/

#define NUM	1
#define ID	2

char saved[256];

void save(int c)
{
    int len = strlen(saved);
    saved[len] = c;
    saved[len+1] = '\0';
}

int lex_type;

int lex_num() { return atoi(saved); }

char *lex_id() { return saved; }

int lex_token(PACKFILE *f)
{
    int c;
    
    saved[0] = '\0';
        
    while (1) {
	c = pack_getc(f);
	if (c == EOF)
	    break;
	
	if (isspace(c)) {
	    if (saved[0])
		break;
	    continue;
	}
	
	if (c == '(') {	
	    while (c != ')' && c != '\n' && c != EOF)
		c = pack_getc(f);
	    continue;
	}

	save(c);
    }

    if (!saved[0])
	return 0;
    
    if (isdigit(saved[0]) || ((saved[0] == '-' || saved[0] == '+')
			      && isdigit(saved[1])))
	return (lex_type = NUM);
    else
	return (lex_type = ID);
}


/*-------------------------------------------------------------------
 	A parser / code generator with many arbitary limits
 -------------------------------------------------------------------*/

int parse1(PACKFILE *f, char *endword1, char *endword2);
int parse(PACKFILE *f, char *endword1, char *endword2);

int error = 0;

int instructions[4096];
int pos = 1;	/* start at 1 so we can use 0 for error checking */

void insert(int i)
{
    if (pos >= (sizeof instructions / sizeof(int))) {
	printf("Out of instruction space.  You nut!\n");
	exit(1);
    }
    
    instructions[pos++] = i;
}

/*-------------------------------------------------------------------*/

struct {
    char id[64];
    int pos;			/* -1 if variable, >= 0 if label */
} symbol_table[200];

int symbol(char *id)
{
    int i;
    
    for (i = 0; symbol_table[i].id[0]; i++)
	if (strcmp(symbol_table[i].id, id) == 0)
	    return i;
    
    strcpy(symbol_table[i].id, id);
    return i;
}
    
/*-------------------------------------------------------------------*/

void _label(PACKFILE *f) 
{
    int s;
    
    if (lex_token(f) != ID) {
	error = 1;
	return;
    }

    s = symbol(lex_id());
    symbol_table[s].pos = pos;
}


void _if(PACKFILE *f) 
{
    int savepos;
    
    savepos = pos; insert(0);	/* place holder for jump dest */
    insert(JUMPF);
    
    /* parse till ELSE or THEN */
    parse(f, "ELSE", "THEN");

    if (strcmp(lex_id(), "ELSE") == 0) {
	int thenpos = pos; insert(0);	/* doesn't assembly suck? */
	insert(JUMP);
	instructions[savepos] = pos;
	savepos = thenpos;
	
	/* parse till THEN */
	parse(f, "THEN", 0);
    }
    
    instructions[savepos] = pos;
}

void _begin(PACKFILE *f)
{
    int beginpos = pos;

    /* parse till UNTIL or WHILE */
    parse(f, "UNTIL", "WHILE");
    
    if (strcmp("UNTIL", lex_id()) == 0) {
	insert(beginpos);
	insert(JUMPF);			/* different to Forth */
    }
    else if (strcmp("WHILE", lex_id()) == 0) {
	int savepos = pos; insert(0);	/* place holder */
	insert(JUMPF);
	
	/* parse till REPEAT */
	parse(f, "REPEAT", 0);
	
	insert(beginpos);
	insert(JUMP);	
	instructions[savepos] = pos;
    }
    else
    	error = 1;
}

void _arrayaccess(PACKFILE *f)
{
    /* syntactical sugar.  This:
	
         [ array element ]

       becomes:

         array @ element @ @@

       and this is allowed:

         [ 2darray elem1 elem11 ]
    */

    int n = 0;
    
    while (1) {
	if (parse1(f, "]", 0) != 0)
	    break;

	if (lex_type == ID) {
	    if (!strcmp(lex_id(), "]")) 
		break;
	
	    insert(GET);
	}
	
	if (n++ > 0)
	    insert(PTRGET);
    }
}

/*-------------------------------------------------------------------*/

typedef void (*proc_t)(PACKFILE *f);

typedef struct {
    char *id;
    int code;
    proc_t proc;
} res_t;
    
res_t reserved_table[] = {
    { "!",	PUT },
    { "@", 	GET },
    { "DUP", 	DUP },
    { "DROP", 	DROP }, 
    { "SWAP", 	SWAP },
    { "OVER",   OVER },
    { "!!", 	PTRPUT },
    { "@@", 	PTRGET },
    { "+",	PLUS },
    { "-",	MINUS },
    { "*",	MULT },
    { "/",	DIV },
    { "=", 	EQ },
    { "<",	LT },
    { "<=",	LTEQ },
    { ">",	GT },
    { ">=",	GTEQ },
    { "AND", 	AND },
    { "OR",	OR },
    { "NOT", 	NOT },
    { "CALL",	CALL },
    { ";", 	RET },
    { ".", 	PRINT },
    { "IF", 	0, _if },
    { "BEGIN",  0, _begin },
    { ":", 	0, _label },
    /* begin syntactical sugar */
    { "[",	0, _arrayaccess },
    { NULL }
};

res_t *find_reserved(char *id)
{
    res_t *r = reserved_table;
    
    while (r->id) {
	if (strcmp(id, r->id) == 0)
	    return r;
	r++;
    }
    
    return NULL;
}

/*-------------------------------------------------------------------*/

int parse1(PACKFILE *f, char *endword1, char *endword2)
{
    char *id;
    res_t *res;
    
    if (!lex_token(f)) 
	return 1;
	
    if (lex_type == NUM) {
	int n = lex_num();
	if (n < 0) {	/* hack */
	    insert(NEGNUM);
	    n = -n;
	}
	insert(n);
	return 0;
    }
	
    if (lex_type != ID) 
	return -1;
    id = lex_id();

    if ((endword1 && !strcmp(id, endword1))
	|| (endword2 && !strcmp(id, endword2)))
	return 1;
    
    res = find_reserved(id);
    if (res) {
	if (res->code)
	    insert(res->code);
	else {	    
	    error = 0;
	    res->proc(f);
	    if (error)
		return -1;
	}
    }
    else 
	insert(symbol(id));

    return 0;
}


int parse(PACKFILE *f, char *endword1, char *endword2)
{
    int x;

    while (1) {
	x = parse1(f, endword1, endword2);
	if (x < 0) 
	    return 0;
	else if (x > 0) 
	    return pos;
    }
}


/*-------------------------------------------------------------------
 	And a nice pretty front end
 -------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
    PACKFILE *f;
    
    install_allegro(SYSTEM_NONE, &errno, atexit);
    
    if (argc != 3) {
	printf("usage: %s input.if output.ifc\n", argv[0]);
	return 1;
    }
    
    f = pack_fopen(argv[1], "r");
    if (!f) {
	printf("Error reading\n");
	return 1;
    }

    parse(f, 0, 0);
    pack_fclose(f);
    
    if (!pos) {
	printf("Error parsing\n");	/* good luck debugging! */
	return 1;
    }
    
    f = pack_fopen(argv[2], "wp");
    if (!f) {
	printf("Error writing\n");
	return 1;
    }
 
    /* write stuff */
    {
	int i;

	/* instructions */
	for (i = 0; i < pos; i++)
	    pack_iputw(instructions[i], f);
	pack_iputw(NOMORE, f);
	
	/* symbol table */
	for (i = 0; symbol_table[i].id[0]; i++) {
	    pack_fputs(symbol_table[i].id, f);
	    pack_putc('\n', f);
	    pack_iputw(symbol_table[i].pos, f);
	}
    }
 
    pack_fclose(f);    
    
    return 0;
}

END_OF_MAIN();
