/*
 * SizeHack 2000 entry
 *
 * Peter Wang <tjaden@psynet.net>
 */

#include <stdlib.h>
#include <time.h>
#include <allegro.h>
#include "vm.h" 


/*
 * Allegro uses an array of chars for `key', which is no good for int
 * Forthish, so we install our own keyboard handler.
 */

volatile int int_key[KEY_MAX];

void int_key_callback(int sc)
{
    int_key[sc & 0x7f] = !(sc & 0x80);
}

END_OF_FUNCTION(int_key_callback);

void install_keyboard_callback()
{
    LOCK_VARIABLE(int_key);
    LOCK_FUNCTION(int_key_callback);
    keyboard_lowlevel_callback = int_key_callback;
}

/*--------------------------------------------------------------------*/

volatile int _time[1] = { 0 };

void speed_regulator()
{
    _time[0]++;
}

END_OF_FUNCTION(speed_regulator);

void install_speed_regulator()
{
    LOCK_VARIABLE(_time);
    LOCK_FUNCTION(speed_regulator);
    install_int_ex(speed_regulator, BPS_TO_TIMER(40));
}

/*--------------------------------------------------------------------*/

void make_vector()
{
    int size, *p;
    size = pop();
    p = malloc(size * sizeof(int));
    push(CVAR(p));
}

/* No room in sizehack for a garbage collector :-) */
void free_vector()
{
    int *p = (int *)pop();
    free(p);
}

/*--------------------------------------------------------------------*/

/* what game would be without one? */
void _rand()
{
    push(rand() % pop());
}

/*--------------------------------------------------------------------*/

void make_bitmap()
{
    int h, w;
    BITMAP *b;
    
    h = pop(); w = pop();
    b = create_bitmap(w, h);
    push(CVAR(b));
}

void free_bitmap()
{
    BITMAP *b = (BITMAP *) pop();
    destroy_bitmap(b);
}

/*--------------------------------------------------------------------*/

void _blit()
{
    BITMAP *d, *s;
    int x, y;
    d = (BITMAP *) pop();
    s = (BITMAP *) pop();
    y = pop(); x = pop();
    blit(s, d, x, y, 0, 0, s->w, s->h);
}

void _clear()
{
    clear((BITMAP *) pop());
}

void __putpixel()
{
    BITMAP *b;
    int x, y, c;
    c = pop(); y = pop(); x = pop();
    b = (BITMAP *) pop();
    putpixel(b, x, y, c);
}

void _circlefill()
{
    BITMAP *b;
    int x, y, r, c;
    c = pop(); r = pop(); y = pop(); x = pop();
    b = (BITMAP *) pop();
    circlefill(b, x, y, r, c);
}

void _putchar()
{
    BITMAP *b;
    int x, y, col;
    char tmp[2] = { 0, 0 };

    col = pop(); tmp[0] = pop();
    y = pop(); x = pop(); 
    b = (BITMAP *) pop();

    text_mode(-1);
    textout(b, font, tmp, x-3, y-3, col);
}

void _putnum()
{
    BITMAP *b;
    int x, y, col, num;

    num = pop();
    col = pop(); 
    y = pop();
    x = pop(); 
    b = (BITMAP *) pop();

    text_mode(-1);
    textprintf_centre(b, font, x, y-3, col, "%d", num);
}

/*--------------------------------------------------------------------*/

void bind_symbols()
{
    static int string[] = { 'S', 'T', 'R', 'I', 'N', 'G' };

    /* cheat a little */
    #define BIND(x, y)	vm_bind_symbol(x, y)

    BIND("time",	CVAR(_time));
        
    BIND("MAKE-VECTOR", CFUNC(make_vector));
    BIND("FREE-VECTOR", CFUNC(free_vector));

    BIND("RAND", 	CFUNC(_rand));
    
    BIND("MAKE-BITMAP", CFUNC(make_bitmap));
    BIND("FREE-BITMAP", CFUNC(free_bitmap));
    
    BIND("BLIT", 	CFUNC(_blit));
    BIND("CLEAR", 	CFUNC(_clear));
    BIND("PUTPIXEL", 	CFUNC(__putpixel));
    BIND("CIRCLEFILL", 	CFUNC(_circlefill));
    BIND("PUTCHAR", 	CFUNC(_putchar));
    BIND("PUTNUM", 	CFUNC(_putnum));
    BIND("screen", 	CVAR(screen));
    
    BIND("string",	CVAR(string));
    
    BIND("key", 	CVAR(&int_key));
    BIND("UP", 		KEY_UP);
    BIND("DOWN", 	KEY_DOWN);
    BIND("LEFT", 	KEY_LEFT);
    BIND("RIGHT", 	KEY_RIGHT);
    BIND("SPACE", 	KEY_SPACE);
    BIND("ESC", 	KEY_ESC);
}

/*--------------------------------------------------------------------*/

int main()
{
    allegro_init();
    install_keyboard();
    install_keyboard_callback();
    install_timer();
    install_speed_regulator();
        
    if (set_gfx_mode(GFX_AUTODETECT, 320, 200, 0, 0) < 0) {
	allegro_message("Error setting video mode (320x200x8bpp)\n");
	return 1;
    }

    srand(time(0));
    
    if (vm_load_file("game.ifc") < 0) {
	allegro_message("Error loading game.ifc\n");
	return 1;
    }
    
    bind_symbols();
    
    vm_execute(vm_symbol_value("main"));
    
    return 0;
}

END_OF_MAIN();
